/*
SendBreak
SockName
*/

/* screader - A screen reader using software TTS
   (c)1997 J. Lemmens
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include "tts.h"
#include "scr.h"

int file_d;
scrstat scr;

/***********************************************************\
* Show_ABT_position - Puts a reverse space on the screen to *
*                     show the position of the ABT.         *
\***********************************************************/

void Show_ABT_position (int y)
{
	int w, attr_on = 0x70, attr_off = 0x07, x;

#ifdef __ABT320
	if (scr.posx == 65)
		return;
	w = open ("/dev/vcsa0", O_RDWR);
	for (x = 0; x <= y; x++)
	{
		lseek (w, 3 + 2 * x * scr.cols, 0);
		write (w, &attr_off, 1);
	} /* if */
	lseek (w, 3 + 2 * (y + 1) * scr.cols, 0);
	write (w, &attr_on, 1);
	for (x = y + 2; x <= scr.rows; x++)
	{
		lseek (w, 3 + 2 * x * scr.cols, 0);
		write (w, &attr_off, 1);
	} /* if */
	close (w);
#endif
} /* Show_ABT_position */

/*********************************************************\
* beep - Make a beep in a certain frequency and duration. *
\*********************************************************/

void beep (int x, int dur)
{
#ifdef linux
	int snd;

	snd = open ("/dev/console", O_WRONLY);
	ioctl (snd, KDMKTONE, (dur << 16) + x);
	close (snd);
#endif
} /* beep */

/*********************************************************************\
* getline - gets a line from file descriptor and returns its pointer. *
\*********************************************************************/

char *getline (FILE *r)
{
	static char hulp[100];
	int byte;

	byte = fgetc (r);
	if (byte == -1)
		return "";
	if (byte == '\n')
		*hulp = 0;
	else
	{
		ungetc (byte, r);
		fscanf (r, "%[^\n]%*c", hulp);
	} /* if */
	return (char *) hulp;
} /* getline */

/*************************************************************\
* Get_TTS_command - Get the TTS-command from the config-file. *
\*************************************************************/

void Get_TTS_command ()
{
	int x = 0;
	static int tts_number = 0;
	static char str[255];
	FILE *r;

	sprintf (str, "%s/%s", abt_dir, TTS_FILE);
	if (! (r = fopen (str, "r")))
	{
#ifdef __SCREADER
		sprintf (abt_dir, "\7%s", str);
		perror (abt_dir);
		exit (1);
#else
		*tts_command = 0;
		return;
#endif
	} /* if */
	*str = 0;
	while (1) /* forever */
	{
		do
		{
			strcpy (str, getline (r));
			if (feof (r))
			{
				rewind (r);
				x = tts_number = 0;
			} /* if */
		} while (*str == '#' || *str == 0);
		if (x++ == tts_number)
		{
			tts_number++;
			break;
		} /* if */
	} /* while */
	fclose (r);
	for (x = 0; str[x] != '#' && str[x] != 0; x++);
	str[x] = 0;
	strcpy (tts_command, str);
} /* Get_TTS-command */

/****\
* Repeat_str - Search repeated chars and substitute with char/n_times. *
\*****/

char *Repeat_str (char *str, int *len)
{
	char *s;
	int l = 0, n, x;

	s = malloc ((scr.rows + 1) * scr.cols);
	bzero (s, (scr.rows + 1) * scr.cols);
	x = -1;
	while (l < *len)
	{
		s[++x] = str[l++];
		if (!isdigit (s[x]) && s[x] != ' ' && s[x] == str[l] && s[x] == str[l + 1])
		{
			n = l;
			while (str[l++] == s[x]);
			sprintf (s, "%s %d times ", s, l-- - n);
			x = strlen (s) - 1;
		} /* if */
	} /* while */
	*len = strlen (s);
	return s;
} /* Repeat_str */

#ifdef __SCREADER
/*******************************************************************\
* Cap_str - Prefix every capitalized character with the word 'cap'. *
\*******************************************************************/

char *Cap_str (char *str, int *len)
{
	char *s;
	int l = 0, n, x;

	s = malloc ((scr.rows + 1) * scr.cols);
	bzero (s, (scr.rows + 1) * scr.cols);
	x = 0;
	while (l < *len)
	{
		if (isupper (str[l]))
		{
			sprintf (s, "%s cap  ", s);
			x = strlen (s) - 1;
		} /* if */
		s[x++] = str[l++];
	} /* while */
	*len = strlen (s);
	return s;
} /* Cap_str */
#endif 

char *Punctuation (char *str, int *len)
{
	char *s;
	int l = 0;

	s = malloc ((scr.rows + 1) * scr.cols);
	bzero (s, (scr.rows + 1) * scr.cols);
   while (l < *len)
   {
      switch (str[l])
      {
      case ',': // comma
         strcat (s, " comma ");
         l++;
         break;
      case ';': // semi colon
         strcat (s, " semi colon ");
         l++;
         break;
      case '.': // point
         strcat (s, " point ");
         l++;
         break;
      case ':': // colon
         strcat (s, " colon ");
         l++;
         break;
      case '-': // dash
         strcat (s, " dash ");
         l++;
         break;
      case '!': // exclamation mark
         strcat (s, " exclamation mark ");
         l++;
         break;
      case '?': // question mark
         strcat (s, " question mark ");
         l++;
         break;
      case '\'': // quote
         strcat (s, " quote ");
         l++;
         break;
      case '"': // double quote
         strcat (s, " double quote ");
         l++;
         break;
      case '(': // open parenthesis
         strcat (s, " open parenthesis ");
         l++;
         break;
      case ')': // close parenthesis
         strcat (s, " close parenthesis ");
         l++;
         break;
      case '[': // open bracket
         strcat (s, " open bracket ");
         l++;
         break;
      case ']': // close bracket
         strcat (s, " close bracket ");
         l++;
         break;
      case '{': // open brace
         strcat (s, " open brace ");
         l++;
         break;
      case '}': // close brace
         strcat (s, " close brace ");
         l++;
         break;
      default:
         s[strlen (s)] = str[l++];
      } // switch
   } /* while */
	*len = strlen (s);
	return s;
} // Punctuation

/*******************************\
* TTS - Text-to-speech handler. *
\*******************************/

pid_t TTS (char *str, int length, int mode)
{
	static pid_t pgid;
	FILE *r, *w, *pipe;

	if (fork ())
		return;

	fclose (stdin);
	fclose (stdout);
	fclose (stderr);
	signal (SIGTERM, SIG_DFL);
	/* kill old speech */
	if (r = fopen (TTS_PID_FILE, "r"))
	{
		fscanf (r, "%d", &pgid);
		fclose (r);
		remove (TTS_PID_FILE);
		killpg (pgid, SIGTERM);

		/* for the festival TTS */
		if (strncmp (tts_command, "festival", 8) == 0)
			system ("killall audsp > /dev/null 2>&1");
	} /* if */

	if (mode == 0)
		raise (SIGTERM);

	/* save pid */
	w = fopen (TTS_PID_FILE, "w");
	fprintf (w, "%d\n", pgid = setsid ());
	fclose (w);

#ifdef __SCREADER
	if (repeat_mode)
#endif
		str = Repeat_str (str, &length);
#ifdef __SCREADER
	if (cap_mode)
		str = Cap_str (str, &length);
#endif
	str = Punctuation (str, &length);
	if (! (pipe = popen (tts_command, "w")))
	{
		remove (TTS_PID_FILE);
		raise (SIGTERM);
	} /* if */
	fwrite (str,  length, 1, pipe);
	fwrite ("\r", 1,      1, pipe);
	pclose (pipe);
	remove (TTS_PID_FILE);
	raise (SIGTERM);
	return pgid;
} /* TTS */

#if defined (__SCREADER) || defined (__XSCREADER)
/***********************************************\
* Speak_Digits - Speak digits instead of value. *
\***********************************************/

void Speak_Digits (char *str, int len)
{
	char hulp[255], *p, *s;

	p = hulp;
	s = str;
	while (len-- > 0)
	{
		*p++ = *s++;
		*p++ = ' ';
	} /* while */
	TTS (hulp, p - hulp, 1);
} /* Speak_Digits */
#endif

/*************************************************\
* Speak_Word - Speak words instead of characters. *
\*************************************************/

void Speak_Word (char *buftts, int ptr)
/* original
{
	int x, p = ptr;

	while (p > 0 && p % scr.cols != 0 && isalnum (buftts[p - 1]))
		p--;
	for (x = p + 1; isalnum (buftts[x]) && x % scr.cols != 0; x++);
#ifdef __SCREADER
	if (speak_value)
#endif
		TTS (buftts + p, x - p, 1);
#ifdef __SCREADER
	else
		Speak_Digits (buftts + p, x - p);
#endif
} / * Speak_Word * /
*/
{
	int x, p = ptr;

	if (! isalnum (buftts[p]))
	{
		TTS (buftts + p, 1, 1);
		return;
	} /* if */
	while (p > 0 && p % scr.cols != 0 && isalnum (buftts[p - 1]))
		p--;
	for (x = p + 1; isalnum (buftts[x]) && x % scr.cols != 0; x++);
#ifdef __SCREADER
	if (speak_value)
#endif
		TTS (buftts + p, x - p, 1);
#ifdef __SCREADER
	else
		Speak_Digits (buftts + p, x - p);
#endif
} /* Speak_Word */

#if defined (__SCREADER) || defined (__XSCREADER)
/***************************************************************\
* Auto_TTS - Reads screen and puts through to TTS if necessary. *
\***************************************************************/

inline void Auto_TTS ()
{
	int p;

	Show_ABT_position (ptr / scr.cols);
	memcpy (old_str, buftts + ptr / scr.cols  * scr.cols, scr.cols);
	old_posx = scr.posx;
	old_posy = scr.posy;
	scr = getstat ();
	if (getscr (0, 0, scr.cols, scr.rows, buftts, SCR_TEXT)   == NULL)
	{
		FILE *w = fopen ("/dev/console", "w");
		beep (TOGGLE_OFF, DURATION);
		fputs ("An buffer-overflow has occurred!\n\r\n\r", w);
		fclose (w);
		raise (SIGQUIT);
	} /* if */
	if (follow_cursor && (old_posx != scr.posx || old_posy != scr.posy))
		ptr = scr.posy * scr.cols + scr.posx;
	if (line_speak && ! key_pressed &&
	    memcmp (old_str, buftts + ptr / scr.cols * scr.cols, scr.cols) != 0)
		TTS (buftts + ptr / scr.cols * scr.cols, scr.cols, 1);
	if (word_mode && (scr.posx != old_posx || scr.posy != old_posy) &&
// fout als cursor regel naar beneden gaat
	    !isalnum (buftts[scr.posy * scr.cols + scr.posx - 1]))
	{
		p   = ptr;
		ptr = old_posy * scr.cols + old_posx - 1;
		Speak_Word (buftts, ptr);
		ptr = p;
	} /* if */
   if (auto_speak && ! word_mode)
   {
      if (old_posx + 1 == scr.posx)
         if (isspace (buftts[scr.posy * scr.cols + scr.posx - 1]))
            TTS ("space", 5, 1);
         else
            TTS (buftts + scr.posy * scr.cols + scr.posx - 1, 1, 1);
      if (old_posx - 1 == scr.posx)
         if (isspace (last_tts_character))
            TTS ("space", 5, 1);
         else
            TTS ((char *) &last_tts_character, 1, 1);
   } // if
	key_pressed = 0;

	/* save last character */
	last_tts_character = buftts[scr.posy * scr.cols + scr.posx - 1];
} /* Auto_TTS */
#endif

#if defined (__SCREADER) || defined (__XSCREADER)
void inline top_of_screen ()
{
	ptr = 0;
	TTS ("top of screen", 13, 1);
} /* top_of_screen */
#endif

#if defined (__SCREADER) || defined (__XSCREADER)
void inline end_of_screen ()
{
	ptr = scr.rows * scr.cols - 1;
	TTS ("end of screen", 13, 1);
} /* end_of_screen */
#endif

#if defined (__SCREADER) || defined (__XSCREADER)
inline void space_left ()
{
	while (buftts[ptr] == ' ')
	{
		if (--ptr < 0)
		{
			top_of_screen ();
			return;
		} /* if */
		if (ptr % scr.cols == scr.cols - 1)
			beep (TOGGLE_ON, DURATION);
	} /* while */
} /* space_left */
#endif

#if defined (__SCREADER) || defined (__XSCREADER)
inline void space_right ()
{
	while (buftts[ptr] == ' ')
	{
		if (++ptr >= scr.rows * scr.cols)
		{
			end_of_screen ();
			return;
		} /* if */
		if (ptr % scr.cols == 0)
			beep (TOGGLE_ON, DURATION);
	} /* while */
} /* space_right */
#endif

#if defined (__SCREADER) || defined (__XSCREADER)
inline void begin_of_word ()
{
	while (isalnum (buftts[ptr - 1]) && ptr % scr.cols != 0)
		if (--ptr < 0)
		{
			top_of_screen ();
			return;
		} /* if */
} /* begin_of_word */
#endif

#if defined (__SCREADER) || defined (__XSCREADER)
inline void end_of_word ()
{
	while (isalnum (buftts[ptr + 1]) && ptr % scr.cols != scr.cols - 1)
		if (++ptr >= scr.rows * scr.cols)
		{
			end_of_screen ();
			return;
		} /* if */
} /* end_of_word */
#endif

#if defined (__SCREADER) || defined (__XSCREADER)
/******************************************\
* Do_TTS_Action - Handle the TTS requests. *
\******************************************/

int Do_TTS_Action (char *ibuf, int ilen)
{
	char hulp[255];
	static int key_click = 0, direct_mode = 0;

	if (key_click)
		beep (2000, 10);
	if (! direct_mode)
		if (ilen != 2 || *ibuf != 0x1b)
			return -1;

	key_pressed = 1;
	switch (ibuf[1 - direct_mode])
	{
/* to build in
speak out attributes
speak out light bars
when number contain a hyphen, speak digits
helpscreen in screader
*/
	case 'a': /* auto_speak */
		auto_speak = 1 - auto_speak;
		if (auto_speak)
			beep (TOGGLE_ON, DURATION);
		else
			beep (TOGGLE_OFF, DURATION);
		break;
	case 'b': /* begin of line */
		ptr = ptr / scr.cols * scr.cols;
		if (line_speak)
			TTS (buftts + ptr /scr.cols * scr.cols, scr.cols, 1);
		break;
	case 'c': /* speak line containing cursor */
		TTS (buftts + scr.posy * scr.cols, scr.cols, 1);
		break;
	case 'd': /* down */
		if ((ptr += scr.cols) >= scr.cols * scr.rows)
		{
			ptr = scr.cols * scr.rows - 1;
			TTS ("end of screen", 13, 1);
		} /* if */
		else
			if (line_speak)
				TTS (buftts + ptr / scr.cols * scr.cols, scr.cols, 1);
		break;
	case 'e': /* speak char/word under cursor */
		if (word_mode)
			Speak_Word (buftts, scr.posy * scr.cols + scr.posx);
		else
			TTS (buftts + scr.posy * scr.cols + scr.posx, 1, 1);
		break;
	case 'f': /* follow cursor */
		if (follow_cursor = 1 - follow_cursor)
			beep (TOGGLE_ON, DURATION);
		else
		{
			beep (TOGGLE_OFF, DURATION);
			Silence ();
		} /* if */
		break;
	case 'g': // bottom
		ptr = (scr.rows - 1) * scr.cols;
		if (line_speak)
			TTS (buftts + ptr, scr.cols, 1);
		break;
	case 'h': /* home */
		ptr = 0;
		if (line_speak)
			TTS (buftts, scr.cols, 1);
		break;
	case 'i': /* line_speak */
		line_speak = 1 - line_speak;
		if (line_speak)
			beep (TOGGLE_ON, DURATION);
		else
			beep (TOGGLE_OFF, DURATION);
		break;
	case 'k': /* key click */
		key_click = 1 - key_click;
		if (key_click)
			beep (TOGGLE_ON, DURATION);
		else
		{
			beep (TOGGLE_OFF, DURATION);
			Silence ();
		} /* if */
		break;
	case ',': /* left one word */
		if (isalnum (buftts[ptr]))
			begin_of_word ();
		if (--ptr < 0)
		{
			top_of_screen ();
			break;
		} /* if */
		if (ptr % scr.cols == scr.cols - 1)
			beep (TOGGLE_ON, DURATION);
		if (buftts[ptr] == ' ')
			space_left ();
		if (isalnum (buftts[ptr]))
			begin_of_word ();
		if (auto_speak)
			if (isalnum (buftts[ptr]))
				Speak_Word (buftts, ptr);
			else
				TTS (buftts + ptr, 1, 1);
		break;
	case 'l': /* left */
		ptr--;
		if (ptr < 0)
		{
			ptr = 0;
			top_of_screen ();
			break;
		} /* if */
		if ((ptr + 1) % scr.cols == 0)
			beep (TOGGLE_OFF, DURATION);
		if (auto_speak)
			TTS (buftts + ptr, 1, 1);
		break;
	case 'm': /* word_mode */
		word_mode = 1 - word_mode;
		if (word_mode)
			beep (TOGGLE_ON, DURATION);
		else
			beep (TOGGLE_OFF, DURATION);
		break;
	case 'n': /* numbers */
		speak_value = 1 - speak_value;
		if (speak_value)
			beep (TOGGLE_ON, DURATION);
		else
			beep (TOGGLE_OFF, DURATION);
		break;
	case 'o': /* silence */
		beep (TOGGLE_OFF, DURATION);
		Silence ();
		break;
	case 'p': /* speak char/word under ptr */
		if (word_mode)
			Speak_Word (buftts, ptr);
		else
			TTS (buftts + ptr, 1, 1);
		break;
	case '.': /* right one word */
		if (isalnum (buftts[ptr]))
			end_of_word ();
		if (++ptr >= scr.rows * scr.cols)
		{
			end_of_screen ();
			break;
		} /* if */
		if (ptr % scr.cols == 0)
			beep (TOGGLE_ON, DURATION);
		if (buftts[ptr] == ' ')
			space_right ();
		if (isalnum (buftts[ptr]))
			begin_of_word ();
		if (auto_speak)
			if (isalnum (buftts[ptr]))
				Speak_Word (buftts, ptr);
			else
				TTS (buftts + ptr, 1, 1);
		break;
	case 'r': /* right */
		ptr++;
		if (ptr >= scr.cols * scr.rows)
		{
			ptr = scr.cols * scr.rows - 1;
			TTS ("end of screen" , 13, 1);
			break;
		} /* if */
		if (ptr % scr.cols == 0)
			beep (TOGGLE_OFF, DURATION);
		if (auto_speak)
			TTS (buftts + ptr, 1, 1);
		break;
	case 's': /* speak line containing ptr */
		TTS (buftts + ptr / scr.cols * scr.cols, scr.cols, 1);
		break;
	case 'u': /* up */
		if ((ptr -= scr.cols) < 0)
		{
			top_of_screen ();
		} /* if */
		else
			if (line_speak)
				TTS (buftts + ptr / scr.cols * scr.cols, scr.cols, 1);
		break;
	case 'w': /* whole screen */
		TTS (buftts, scr.rows * scr.cols, 1);
		break;
	case 'x': /* other tts */
		beep (TOGGLE_ON, DURATION);
		Get_TTS_command ();
		sprintf (hulp, "This is the %s speech synthesizer", &tts_command);
		TTS (hulp, strlen (hulp), 1);
		break;
	case 'z': /* suspend / direct mode */
		direct_mode = 1 - direct_mode;
		if (direct_mode)
			beep (TOGGLE_ON, DURATION);
		else
			beep (TOGGLE_OFF, DURATION);
		break;
	case '-': /* repeat_mode */
		repeat_mode = 1 - repeat_mode;
		if (repeat_mode)
			beep (TOGGLE_ON, DURATION);
		else
			beep (TOGGLE_OFF, DURATION);
		break;
	case '\'': /* cap_mode */
		cap_mode = 1 - cap_mode;
		if (cap_mode)
			beep (TOGGLE_ON, DURATION);
		else
			beep (TOGGLE_OFF, DURATION);
		break;
	case '=': /* ptr = cursor */
		ptr = scr.posy * scr.cols + scr.posx;
		break;
	default:
// logger (" ibuf: %d", ibuf[1]);
		return -1;
	} /* switch */
	return 0;
} /* Do_TTS_Action */
#endif

#if defined (__SCREADER) || defined (__XSCREADER)
/***********************************************\
* Init_TTS - Initialize everything for the tts. *
\***********************************************/

void Init_TTS ()
{
	int d, kdmode;
	char *kdmsg = "Screader only works on virtual consoles.";
	struct stat st;
	FILE *r;

	strcpy (abt_dir, "/opt/etc/abt320");
	stat ("/usr/lib/terminfo/v/vt100", &st);
	tc_size = st.st_size;
	termcap = malloc (tc_size);
	r = fopen ("/usr/lib/terminfo/v/vt100", "rb");
	fread (termcap, tc_size, 1, r);
	fclose (r);
	getscrinfo ();
	if ((buftts   = malloc ((scr.rows + 2) * scr.cols + 5)) == NULL)
	{
		FILE *w = fopen ("/dev/console", "w");
		beep (TOGGLE_OFF, DURATION);
		fprintf (w, "Can't allocate enough memory.\n\r");
		fclose (w);
		raise (SIGKILL);
	} /* if */
	ptr = scr.posy;
	Get_TTS_command ();
	d = open ("/dev/console", O_RDONLY);
	ioctl (d, KDGETMODE, &kdmode);
	close (d);
} /* Init_TTS */
#endif

